/*
 * Copyright (c) 2020 - 2021 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __UI_APP_AMBIENT__
#define __UI_APP_AMBIENT__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>

#include <bundle.h>

/**
 * @brief Enumeration for period.
 * @remark This function only for internal applications
 */
typedef enum {
	UI_APP_AMBIENT_UPDATE_NONE,
	UI_APP_AMBIENT_UPDATE_MINUTE,
	UI_APP_AMBIENT_UPDATE_FIVE_MINUTES,
	UI_APP_AMBIENT_UPDATE_FIFTEEN_MINUTES,
	UI_APP_AMBIENT_UPDATE_THIRTY_MINUTES
} ui_app_ambient_update_period_e;

/**
 * @brief Enumeration for event.
 * @remark This function only for internal applications
 */
typedef enum {
	UI_APP_AMBIENT_EVENT_PREPARE,
	UI_APP_AMBIENT_EVENT_READY
} ui_app_ambient_event_e;

/**
 * @brief Enumeration for ambient state.
 * @remark This function only for internal applications
 */
typedef enum {
	UI_APP_AMBIENT_STATE_LEAVE,
	UI_APP_AMBIENT_STATE_ENTER,
	UI_APP_AMBIENT_STATE_BLOCK_TICK
} ui_app_ambient_state_e;

/**
 * @brief Called with the period set with ui_app_ambient_set_update_period() if the device is in the ambient mode.
 * @remark This function only for internal applications
 * @param[in] user_data The user data to be passed to the callback functions
 * @see ui_app_ambient_set_lifecycle()
 * @see ui_app_ambient_set_update_period()
 */
typedef void (*ui_app_ambient_update_frame_cb) (void *user_data);

/**
 * @brief Called when the device enters or exits the ambient mode.
 * @remark This function only for internal applications
 * @param[in] state The ambient mode
 * @param[in] data The extra bundle data from ambient viewer
 * @param[in] user_data The user data to be passed to the callback functions
 * @see ui_app_ambient_set_lifecycle()
 */
typedef void (*ui_app_ambient_changed_cb) (ui_app_ambient_state_e state, bundle *data, void *user_data);

/**
 * @brief The structure type containing the set of callback functions for application events handle.
 * @remark This function only for internal applications
 * @see ui_app_ambient_set_lifecycle()
 */
typedef struct {
	ui_app_ambient_update_frame_cb frame_updated;
	ui_app_ambient_changed_cb ambient_changed;
} ui_app_ambient_lifecycle_callback_s;

/**
 * @brief Sets lifecycle for UI ambient lifecycle.
 * @remark This function only for internal applications
 * @param[in] lifecycle The set of callback functions to handle application lifecycle events
 * @param[in] user_data The user data to be passed to the callback functions
 * @return @c 0 on success,
 *         otherwise a negative error value
 * @see ui_app_ambient_unset_lifecycle()
 */
int ui_app_ambient_set_lifecycle(ui_app_ambient_lifecycle_callback_s *lifecycle, void *user_data);

/**
 * @brief Unsets lifecycle for UI ambient lifecycle.
 * @remark This function only for internal applications
 * @see ui_app_ambient_set_lifecycle()
 */
void ui_app_ambient_unset_lifecycle(void);

/**
 * @brief Sets the period of ui_app_ambient_update_frame_cb() calls.
 * @remark This function only for internal applications
 * @param[in] period period type
 * @return @c 0 on success,
 *         otherwise a negative error value
 * @see ui_app_ambient_get_update_period()
 */
int ui_app_ambient_set_update_period(ui_app_ambient_update_period_e period);

/**
 * @brief Gets the period of ui_app_ambient_update_frame_cb() calls.
 * @remark This function only for internal applications
 * @param[out] period period type
 * @return @c 0 on success,
 *         otherwise a negative error value
 * @see ui_app_ambient_set_update_period()
 */
int ui_app_ambient_get_update_period(ui_app_ambient_update_period_e *period);

/**
 * @brief Notifies event with bundle data to the @a receiver.
 * @remark This function only for internal applications
 * @param[in] event ambient event type
 * @param[in] extra extra bundle data
 * @return @c 0 on success,
 *         otherwise a negative error value
 */
int ui_app_ambient_notify_event(ui_app_ambient_event_e event, bundle *extra);

#ifdef __cplusplus
}
#endif

#endif /* __UI_APP_AMBIENT__ */
