/*
 * Copyright (c) 2020 - 2021 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <alarm.h>
#include <app_common_internal.h>
#include <aul.h>
#include <aul_app_com.h>
#include <bundle_cpp.h>
#include <bundle_internal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#include <memory>
#include <string>

#include "common_private.hh"
#include "log_private.hh"
#include "ui_app_ambient.h"

namespace {

constexpr const int SECOND = 1;
constexpr const int ONE_MINUTE_IN_SEC = 60;
constexpr const int MILLION = 1000000;  // for calculating micro seconds

struct AmbientTickTypeInfo {
  int interval_;
  int minute_base_;
  int hour_base_;
};

AmbientTickTypeInfo ambient_tick_type_infos[] = {
  { 0, 0 , 0},
  { ONE_MINUTE_IN_SEC, 1, 0 },
  { ONE_MINUTE_IN_SEC * 5, 5, 0 },
  { ONE_MINUTE_IN_SEC * 15, 15, 0 },
  { ONE_MINUTE_IN_SEC * 30, 30, 0 }
};

class UIAppAmbient {
 public:
  ~UIAppAmbient() {}

  int SetLifecycleCallback(ui_app_ambient_lifecycle_callback_s* callback,
      void* user_data) {
    if (initialized_)
      return 0;

    callback_ = *callback;
    user_data_ = user_data;

    int ret = aul_app_com_create("ui.ambientchange", nullptr, OnAppComCb, this,
        &conn_);
    if (ret != AUL_R_OK) {
      _E("aul_app_com_create() is failed. error(%d)", ret);
      return ret;
    }

    initialized_ = true;
    return 0;
  }

  void UnsetLifecycleCallback() {
    if (conn_) {
      aul_app_com_leave(conn_);
      conn_ = nullptr;
    }

    if (alarm_id_) {
      alarmmgr_remove_alarm(alarm_id_);
      alarm_id_ = 0;
    }

    callback_.ambient_changed = nullptr;
    callback_.frame_updated = nullptr;
    user_data_ = nullptr;
    initialized_ = false;
  }

  int SetUpdatePeriod(int period_type) {
    period_type_ = period_type;
    if (IsAmbientMode()) {
      if (alarm_id_) {
        alarmmgr_remove_alarm(alarm_id_);
        alarm_id_ = 0;
      }

      if (period_type != UI_APP_AMBIENT_UPDATE_NONE)
        return SetAmbientTick();
    }

    return 0;
  }

  int GetUpdatePeriod() {
    return period_type_;
  }

  int SendAmbientEvent(ui_app_ambient_event_e event, bundle* extra) {
    int ret = SetAppId();
    if (ret != 0)
      return ret;

    tizen_base::Bundle envelope(extra ? extra : bundle_create(), false, true);
    envelope.Add("__APP_AMBIENT_EVENT__", std::to_string(event).c_str());
    envelope.Add("__APP_AMBIENT_SENDER__", app_id_.c_str());
    ret = aul_app_com_send("aod.ambientevent", envelope.GetHandle());
    if (ret != AUL_R_OK) {
      _E("aul_app_com_send() is failed. error(%d)", ret);
      return -1;
    }

    _I("Send AOD Event(%d) done", event);
    return 0;
  }

 private:
  static int OnAppComCb(const char* endpoint, aul_app_com_result_e result,
      bundle* envelope, void* user_data) {
    _E("OnAppComMessageReceived");
    const char* mode = bundle_get_val(envelope, "__AMBIENT_MODE__");
    if (mode == nullptr)
      return -1;

    ui_app_ambient_state_e state = static_cast<ui_app_ambient_state_e>(
        std::stoi(mode));

    const char* extra = bundle_get_val(envelope, "__AMBIENT_EXTRA__");
    if (extra == nullptr)
      return -1;

    tizen_base::Bundle b(extra);

    auto* handle = static_cast<UIAppAmbient*>(user_data);
    auto callback = handle->callback_;
    if (callback.ambient_changed)
      callback.ambient_changed(state, b.GetHandle(), handle->user_data_);

    if (state == UI_APP_AMBIENT_STATE_ENTER && !handle->IsAmbientMode()) {
      int ret = handle->SetAmbientTick();
      if (ret != 0)
        _E("Failed to set alarm");
    } else if (state == UI_APP_AMBIENT_STATE_LEAVE && handle->IsAmbientMode()) {
      handle->UnsetAmbientTick();
    }

    return 0;
  }

  static int AlarmCb(alarm_id_t id, void* user_data) {
    _D("OnAmbientTick");
    auto* handle = static_cast<UIAppAmbient*>(user_data);
    if (handle->alarm_id_ == id) {
      auto callback = handle->callback_;
      callback.frame_updated(handle->user_data_);
    }

    return 0;
  }

  int SetAppId() {
    if (app_id_.empty()) {
      char appid[256] = { 0, };
      int ret = aul_app_get_appid_bypid(getpid(), appid, sizeof(appid));
      if (ret != AUL_R_OK) {
        _E("aul_app_get_appid_bypid() is failed. error(%d)", ret);
        return ret;
      }

      app_id_ = std::string(appid);
    }
    return 0;
  }

  alarm_entry_t* GetAmbientTickAlarm(int period_type) {
    alarm_entry_t* alarm_info = alarmmgr_create_alarm();
    if (alarm_info == nullptr)
      return nullptr;

    std::unique_ptr<alarm_entry_t, decltype(alarmmgr_free_alarm)*> ptr(
        alarm_info, alarmmgr_free_alarm);

    struct timespec current_time;
    clock_gettime(CLOCK_REALTIME, &current_time);
    tzset();
    struct tm current_tm;
    localtime_r(&current_time.tv_sec, &current_tm);

    int offset = 0;
    if (period_type == UI_APP_AMBIENT_UPDATE_FIVE_MINUTES ||
        period_type == UI_APP_AMBIENT_UPDATE_FIFTEEN_MINUTES ||
        period_type == UI_APP_AMBIENT_UPDATE_THIRTY_MINUTES) {
      int minute_base = ambient_tick_type_infos[period_type].minute_base_;
      int remain_min = minute_base - 1 - (current_tm.tm_min % minute_base);
      offset = (remain_min * ONE_MINUTE_IN_SEC) +
        (ONE_MINUTE_IN_SEC - current_tm.tm_sec);
    } else if (period_type == UI_APP_AMBIENT_UPDATE_MINUTE) {
      offset = ONE_MINUTE_IN_SEC - current_tm.tm_sec;
    }

    struct tm next_tm;
    time_t next = current_time.tv_sec + offset;
    localtime_r(&next, &next_tm);

    alarm_date_t date = { 0, };
    date.year = next_tm.tm_year + 1900;
    date.month = next_tm.tm_mon + 1;
    date.day = next_tm.tm_mday;
    date.hour = next_tm.tm_hour;
    date.min = next_tm.tm_min;
    date.sec = next_tm.tm_sec;

    alarmmgr_set_type(alarm_info, ALARM_TYPE_VOLATILE);

    int ret = alarmmgr_set_time(alarm_info, date);
    if (ret != ALARMMGR_RESULT_SUCCESS) {
      _E("alarmmgr_set_time() is failed. error(%d)", ret);
      return nullptr;
    }

    ret = alarmmgr_set_repeat_mode(alarm_info, ALARM_REPEAT_MODE_REPEAT,
        ambient_tick_type_infos[period_type].interval_);
    if (ret != ALARMMGR_RESULT_SUCCESS) {
      _E("alarmmgr_set_repeat_mode() is failed. error(%d)", ret);
      return nullptr;
    }

    LOGD("Next alarm tick [%d:%d:%d]", date.hour, date.min, date.sec);
    return ptr.release();
  }

  int SetAmbientTick() {
    if (period_type_ == UI_APP_AMBIENT_UPDATE_NONE && alarm_id_ == 0) {
      _I("Default period is set [FIFTEEN_MINUTES]");
      period_type_ = UI_APP_AMBIENT_UPDATE_FIFTEEN_MINUTES;
    }

    int ret = SetAppId();
    if (ret != 0)
      return ret;

    ret = alarmmgr_init(app_id_.c_str());
    if (ret != ALARMMGR_RESULT_SUCCESS)
      _E("alarmmgr_init() is failed. error(%d)", ret);

    alarm_entry_t* alarm_info = GetAmbientTickAlarm(period_type_);
    if (alarm_info == nullptr)
      return -1;

    ret = alarmmgr_add_alarm_withcb_with_localtime(alarm_info, AlarmCb,
        this, &alarm_id_);
    if (ret < 0) {
      _E("alarmmgr_add_alarm_withcb_with_localtime() is failed. error(%d)",
          ret);
    }

    callback_.frame_updated(user_data_);
    alarmmgr_free_alarm(alarm_info);
    ambient_mode_ = true;
    return 0;
  }

  void UnsetAmbientTick() {
    if (alarm_id_) {
      alarmmgr_remove_alarm(alarm_id_);
      alarm_id_ = 0;
    }

    ambient_mode_ = false;
  }

  bool IsAmbientMode() {
    return ambient_mode_;
  }

 private:
  bool initialized_ = false;
  std::string app_id_;
  aul_app_com_connection_h conn_ = nullptr;
  ui_app_ambient_lifecycle_callback_s callback_ = { 0, };
  void* user_data_ = nullptr;
  bool ambient_mode_ = false;
  int period_type_ = UI_APP_AMBIENT_UPDATE_NONE;
  int alarm_id_ = 0;
};

UIAppAmbient context;

}  // namespace

API int ui_app_ambient_set_lifecycle(
    ui_app_ambient_lifecycle_callback_s* lifecycle, void* user_data) {
  if (lifecycle == nullptr ||
      lifecycle->ambient_changed == nullptr ||
      lifecycle->frame_updated == nullptr) {
    _E("Invalid parameter");
    return -1;
  }

  return context.SetLifecycleCallback(lifecycle, user_data);
}

API void ui_app_ambient_unset_lifecycle(void) {
  context.UnsetLifecycleCallback();
}

API int ui_app_ambient_set_update_period(
    ui_app_ambient_update_period_e period) {
  _D("Period: %d", period);
  return context.SetUpdatePeriod(period);
}

API int ui_app_ambient_get_update_period(
    ui_app_ambient_update_period_e* period) {
  if (period == nullptr) {
    _E("Invalid parameter");
    return -1;
  }

  *period = static_cast<ui_app_ambient_update_period_e>(
      context.GetUpdatePeriod());
  return 0;
}

API int ui_app_ambient_notify_event(ui_app_ambient_event_e event,
    bundle* extra) {
  return context.SendAmbientEvent(event, extra);
}
