/*
 * Copyright (c) 2011 - 2021 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <app_common_internal.h>

#include "app.h"
#include "app_extension.h"
#include "app_internal.h"
#include "common_private.hh"
#include "log_private.hh"

namespace {

class AppMain {
 public:
  AppMain(app_event_callback_s* cb, void* data)
      : cb_(cb), data_(data) {
  }

  int Run(int argc, char** argv) {
    RegisterSystemEvents();

    ui_app_lifecycle_callback_s callback = {
      .create = OnCreate,
      .terminate = OnTerminate,
      .pause = OnPause,
      .resume = OnResume,
      .app_control = OnAppControl
    };
    int ret = ui_app_main(argc, argv, &callback, this);

    DeregisterSystemEvents();
    return ret;
  }

 private:
  void RegisterSystemEvents() {
    if (cb_->low_memory) {
      ui_app_add_event_handler(&low_memory_,
          APP_EVENT_LOW_MEMORY, OnLowMemory, this);
    }

    if (cb_->low_battery) {
      ui_app_add_event_handler(&low_battery_,
          APP_EVENT_LOW_BATTERY, OnLowBattery, this);
    }

    if (cb_->language_changed) {
      ui_app_add_event_handler(&language_changed_,
          APP_EVENT_LANGUAGE_CHANGED, OnLanguageChanged, this);
    }

    if (cb_->device_orientation) {
      ui_app_add_event_handler(&device_orientation_changed_,
          APP_EVENT_DEVICE_ORIENTATION_CHANGED, OnDeviceOrientationChanged,
          this);
    }

    if (cb_->region_format_changed) {
      ui_app_add_event_handler(&region_format_changed_,
          APP_EVENT_REGION_FORMAT_CHANGED, OnRegionFormatChanged, this);
    }
  }

  void DeregisterSystemEvents() {
    if (region_format_changed_) {
      ui_app_remove_event_handler(region_format_changed_);
      region_format_changed_ = nullptr;
    }

    if (device_orientation_changed_) {
      ui_app_remove_event_handler(device_orientation_changed_);
      device_orientation_changed_ = nullptr;
    }

    if (language_changed_) {
      ui_app_remove_event_handler(language_changed_);
      language_changed_ = nullptr;
    }

    if (low_battery_) {
      ui_app_remove_event_handler(low_battery_);
      low_battery_ = nullptr;
    }

    if (low_memory_) {
      ui_app_remove_event_handler(low_memory_);
      low_memory_ = nullptr;
    }
  }

  static bool OnCreate(void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->create)
      return context->cb_->create(context->data_);

    return false;
  }

  static void OnTerminate(void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->terminate)
      context->cb_->terminate(context->cb_);
  }

  static void OnPause(void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->pause)
      context->cb_->pause(context->data_);
  }

  static void OnResume(void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->resume)
      context->cb_->resume(context->data_);
  }

  static void OnAppControl(app_control_h app_control, void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->app_control)
      context->cb_->app_control(app_control, context->data_);
  }

  static void OnLowMemory(app_event_info_h event_info, void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->low_memory)
      context->cb_->low_memory(context->data_);
  }

  static void OnLowBattery(app_event_info_h event_info, void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->low_battery)
      context->cb_->low_battery(context->data_);
  }

  static void OnDeviceOrientationChanged(app_event_info_h event_info,
      void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->device_orientation) {
      app_device_orientation_e value;
      app_event_get_device_orientation(event_info, &value);
      context->cb_->device_orientation(value, context->data_);
    }
  }

  static void OnLanguageChanged(app_event_info_h event_info, void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->language_changed)
      context->cb_->language_changed(context->data_);
  }

  static void OnRegionFormatChanged(app_event_info_h event_info, void* data) {
    auto* context = static_cast<AppMain*>(data);
    if (context->cb_->region_format_changed)
      context->cb_->region_format_changed(context->data_);
  }

 private:
  app_event_callback_s* cb_;
  void* data_;
  app_event_handler_h low_memory_ = nullptr;
  app_event_handler_h low_battery_ = nullptr;
  app_event_handler_h language_changed_ = nullptr;
  app_event_handler_h device_orientation_changed_ = nullptr;
  app_event_handler_h region_format_changed_ = nullptr;
};

}  // namespace

API int app_main(int argc, char** argv, app_event_callback_s* callback,
    void* data) {
  if (callback == nullptr) {
    _E("Invalid parameter");
    return APP_ERROR_INVALID_PARAMETER;
  }

  AppMain app(callback, data);
  return app.Run(argc, argv);
}

API int app_efl_main(int* argc, char*** argv, app_event_callback_s* callback,
    void* user_data) {
  if (argc == nullptr || argv == nullptr) {
    _E("Invalid parameter");
    return APP_ERROR_INVALID_PARAMETER;
  }

  return app_main(*argc, *argv, callback, user_data);
}

API void app_exit(void) {
  ui_app_exit();
}

API void app_efl_exit(void) {
  app_exit();
}
